
from arkitekt.actors.functional import FunctionalFuncActor
from herre.herre import get_current_herre
from arkitekt.threadvars import get_current_assign
from qtpy.QtCore import QObject, Signal
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from herre.wards.base import WardException
from arkitekt.actors.actify import actify, define
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.base import MessageDataModel, MessageModel
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.agents.base import Agent, AgentException, parse_params
import logging
from herre.console import get_current_console
import uuid
from koil import get_current_koil
from koil.koil import Koil
from koil.qt import FutureWrapper, TimeoutFutureWrapper

logger = logging.getLogger(__name__)




class QtOldActor(QObject):
    provideInSignal = Signal(str, BouncedProvideMessage)
    provideOutSignal = Signal(str)
    provideErrorSignal = Signal(str, Exception)

    assignInSignal = Signal(str,tuple, dict)
    assignOutSignal =  Signal(str, tuple)
    assignErrorSignal = Signal(str, Exception)

    unprovideInSignal = Signal(str, BouncedProvideMessage)
    unprovideOutSignal = Signal(str)
    unprovideErrorSignal = Signal(str, Exception)

    def __init__(self, *args,  qt_assign=None, loop= None, qt_on_provide = None, qt_on_unprovide = None, timeout=300, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.futureMap = {}
        self.koil = get_current_koil()
        self.loop = loop or self.koil.loop

        self.qt_assign = qt_assign
        self.qt_on_provide = qt_on_provide
        self.qt_on_unprovide = qt_on_unprovide

        self.assign_timeout = timeout
        self.provide_timeout = timeout
        self.unprovide_timeout = timeout



        if self.qt_assign is not None:
            self.assignInSignal.connect(self.assign_function_in_qtloop) # insignal will be set dymaically by qt agent
            self.assignOutSignal.connect(self.set_assign_future_qtloop)
            self.assignErrorSignal.connect(self.set_error_future_qtloop)

        if self.qt_on_provide is not None:
            
            self.provideInSignal.connect(self.provide_function_in_qtloop)
            self.provideOutSignal.connect(self.set_provide_future_qtloop)
            self.provideErrorSignal.connect(self.set_error_future_qtloop)

        if self.qt_on_unprovide is not None:
            self.unprovideInSignal.connect(self.unprovide_function_in_qtloop)
            self.unprovideOutSignal.connect(self.set_unprovide_future_qtloop)
            self.unprovideErrorSignal.connect(self.set_error_future_qtloop)

        

    def provide_function_in_qtloop(self, reference, *args, **kwargs):
        try:
            returns = self.qt_on_provide(*args, **kwargs) if self.qt_on_provide else None
            self.provideOutSignal.emit(reference)
        except Exception as e:
            self.provideErrorSignal.emit(reference, e)

    def unprovide_function_in_qtloop(self, reference, *args, **kwargs):
        try:
            returns = self.qt_on_unprovide(*args, **kwargs) if self.qt_on_unprovide else None
            self.unprovideOutSignal.emit(reference)
        except Exception as e:
            self.unprovideErrorSignal.emit(reference, e)

    def set_provide_future_qtloop(self, reference):
        self.loop.call_soon_threadsafe(self.futureMap[reference].set_result, None)

    def set_unprovide_future_qtloop(self, reference):
        self.loop.call_soon_threadsafe(self.futureMap[reference].set_result, None)
    
    def set_assign_future_qtloop(self, reference, returns):
        if len(returns) == 1:
            print("Setting None")
            self.loop.call_soon_threadsafe(self.futureMap[reference].set_result, returns[0])
        if len(returns) >= 2:
            print("Setting Result")
            self.loop.call_soon_threadsafe(self.futureMap[reference].set_result, returns)


    def assign_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_assign(*args, **kwargs)
            if isinstance(returns, tuple):
                self.assignOutSignal.emit(reference, returns)
            else:
                self.assignOutSignal.emit(reference, (returns,))
        except Exception as e:
            self.assignErrorSignal.emit(reference, e)


    def set_error_future_qtloop(self, reference, exception):
        self.loop.call_soon_threadsafe(self.futureMap[reference].set_exception, exception)
    

    async def on_provide(self, message):
        if not self.qt_on_provide: return None
        print(self, message)
        reference = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futureMap[reference] = future
        self.provideInSignal.emit(reference, message)
        return await asyncio.wait_for(future, timeout=self.provide_timeout)

    async def on_unprovide(self,message):
        if not self.qt_on_unprovide: return None
        reference = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futureMap[reference] = future
        self.unprovideInSignal.emit(reference,message)
        return await asyncio.wait_for(future, timeout=self.unprovide_timeout)

    async def assign(self, *args, **kwargs):
        reference = str(uuid.uuid4())
        future = self.loop.create_future()
        self.futureMap[reference] = future
        self.assignInSignal.emit(reference, args, kwargs)

        return await asyncio.wait_for(future, timeout=self.assign_timeout)



class ActorSignals(QObject):

        def __init__(self, timeout=300) -> None:
            self.assign = TimeoutFutureWrapper(timeout=timeout)
            self.on_provide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)
            self.on_unprovide = TimeoutFutureWrapper(timeout=timeout, pass_through=True)



class QtActor(FunctionalFuncActor, QObject):

    def __init__(self, *args,  qt_assign=None, loop= None, qt_on_provide = None, qt_on_unprovide = None, timeout=300, koil: Koil = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        print("oisndofinsofinosienfosinefoisnefoineofinsoienf")
        self.koil = koil or get_current_koil()
        self.loop = self.koil.loop

        self.qt_assign = qt_assign
        self.qt_on_provide = qt_on_provide
        self.qt_on_unprovide = qt_on_unprovide

        self.signals = ActorSignals()

        if self.qt_assign is not None:
            self.signals.assign.call.connect(self.assign_function_in_qtloop) # insignal will be set dymaically by qt agent

        if self.qt_on_provide is not None:
            self.signals.on_provide.connect(self.provide_function_in_qtloop)

        if self.qt_on_unprovide is not None:
            self.signals.on_unprovide.connect(self.unprovide_function_in_qtloop)

        

    def provide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_on_provide(*args, **kwargs) if self.qt_on_provide else None
            self.signals.on_provide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_provide.reject.emit(reference, e)

    def unprovide_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_on_unprovide(*args, **kwargs) if self.qt_on_unprovide else None
            self.signals.on_unprovide.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.on_unprovide.reject.emit(reference, e)


    def assign_function_in_qtloop(self, reference, args, kwargs):
        try:
            returns = self.qt_assign(*args, **kwargs)
            self.signals.assign.resolve.emit(reference, returns)
        except Exception as e:
            self.signals.assign.reject.emit(reference, e)


    async def assign(self, *args, **kwargs):
        print("Calledsssssss")
        await self.signals.assign.acall(*args, **kwargs)

    async def on_provide(self, *args, **kwargs):
        print("Calledssssss2e23e2e1s")
        await self.signals.on_provide.acall(*args, **kwargs)

    async def on_unprovide(self, *args, **kwargs):
        print("Calledssssdsdsdssss")
        await self.signals.on_unprovide.acall(*args, **kwargs)


    




from typing import Dict
from .base import Transport
from arkitekt.config import TransportProtocol


class TransportRegistry:

    def __init__(self) -> None:
        self.registeredTransports: Dict[str, Transport] = {}

    def register_transport(self, protocol: TransportProtocol,  transport: Transport):
        self.registeredTransports[protocol] = transport

    def get_transport_for_protocol(self,protocol):
        return self.registeredTransports[protocol]

    
def register_transport(protocol: TransportProtocol):
    print("Registering Transport")

    def rea_decorator(grant):
        assert issubclass(grant, Transport), "Transport must subclass Transport"
        get_current_transport_registry().register_transport(protocol, grant)
        return grant

    return rea_decorator



GRANT_REGISTRY = None


def get_current_transport_registry(with_defaults=True):
    global GRANT_REGISTRY
    if not GRANT_REGISTRY:
        GRANT_REGISTRY = TransportRegistry()
        if with_defaults:
            import arkitekt.transport.websocket 


    return GRANT_REGISTRY



from ..log import LogDataModel
from ....messages.types import UNASSIGN_LOG
from ....messages.base import MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = UNASSIGN_LOG
    extensions: Optional[MetaExtensionsModel]


class UnassignLogMessage(MessageModel):
    data: LogDataModel
    meta: MetaModel


from pydantic.main import BaseModel
from ...messages.types import AGENT_DISCONNECT
from ...messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = AGENT_DISCONNECT
    extensions: Optional[MetaExtensionsModel]

class DataModel(MessageDataModel):
    canceller: Optional[str] # A reference to the cancelling agent

    
class AgentDisconnectMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from qtpy import QtWidgets


class SettingsPopup(QtWidgets.QDialog):
    def __init__(self, magic_bar, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.magic_bar = magic_bar

        self.layout = QtWidgets.QVBoxLayout()

        self.delete_fakts = QtWidgets.QPushButton("Delete Configuration")
        self.delete_fakts.clicked.connect(self.fakts_delete)

        self.logout_button = QtWidgets.QPushButton("Logout")
        self.logout_button.clicked.connect(self.logout_pressed)

        self.layout.addWidget(self.delete_fakts)
        self.layout.addWidget(self.logout_button)
        self.setLayout(self.layout)

    def fakts_delete(self):
        self.magic_bar.fakts.delete()


    def logout_pressed(self):
        self.magic_bar.herre.logout()



import threading
from typing import Any, Coroutine
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
import contextvars
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.assign.assign_return import AssignReturnMessage
from arkitekt.messages.postman.assign.assign_yield import AssignYieldsMessage
from arkitekt.messages.postman.assign.assign_done import AssignDoneMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.threadvars import assign_message, transport, janus_queue
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.actors.base import Actor
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import asyncio
from arkitekt.packers.utils import shrink_inputs, shrink_outputs, expand_inputs, expand_outputs
import logging
import janus


logger = logging.getLogger(__name__)


class FunctionalActor(Actor):
    pass


    def __init__(self, *args, assign=None, on_provide=None, on_unprovide=None, **kwargs) -> None:
        print(assign)
        super().__init__(*args, **kwargs)
        self.assign = assign or self.assign
        print(self.assign)
        self.on_provide = on_provide or self.on_provide
        self.on_unprovide = on_unprovide or self.on_unprovide


    def assign(self, *args, **kwargs):
        raise NotImplementedError("Please overwrite this")



class FunctionalFuncActor(FunctionalActor):
    assign: Coroutine[Any, Any, Any]

    
    async def progress(self, value, percentage):
        await self._progress(value, percentage)


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            transport.set(self.transport)
            assign_message.set(message)


            returns = await self.assign(*args, **kwargs)


            assign_message.set(None)
            transport.set(None)
            await self.transport.forward(AssignReturnMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))




class FunctionalGenActor(FunctionalActor):


    async def progress(self, value, percentage):
        await self._progress(value, percentage)

    async def on_assign(self, message: BouncedForwardedAssignMessage):
        try:
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            
            transport.set(self.transport)
            assign_message.set(message)

            async for returns in self.assign(*args, **kwargs):
                print("Result of", returns)



                await self.transport.forward(AssignYieldsMessage(data={
                    "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
                }, meta = {
                    "reference": message.meta.reference,
                    "extensions": message.meta.extensions
                }))


            assign_message.set(None)
            transport.set(None)


            await self.transport.forward(AssignDoneMessage(data={
                "returns": await shrink_outputs(self.template.node, returns) if self.shrink_outputs else returns
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))


        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            self.console.print_exception()
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



class FunctionalThreadedFuncActor(FunctionalActor):

    def __init__(self, *args, nworkers=4, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(nworkers)

    async def iterate_queue(self, async_q, message: BouncedForwardedAssignMessage):
        while True:
            val = await async_q.get()
            action = val[0]
            value = val[1]

            if action == "log":
                message = AssignLogMessage(
                    data = {
                        "message": value[0],
                        "level": value[1]
                    },
                    meta = {**message.meta.dict(exclude={"type"})}
                )

                await self.transport.forward(message)
                async_q.task_done()
            if action == "return":
                await self.transport.forward(AssignReturnMessage(data={
                    "returns": await shrink_outputs(self.template.node, value) if self.shrink_outputs else value
                }, meta = {**message.meta.dict(exclude={"type"})}
                ))
                async_q.task_done()
                break
            if action == "exception":
                await self.transport.forward(AssignCriticalMessage(data={
                        "type": val.__class__.__name__,
                        "message": str(val)
                    }, meta = {**message.meta.dict(exclude={"type"})}
                    ))
                async_q.task_done()
                break  



    def _assign_threaded(self, queue, message, args, kwargs):
        janus_queue.set(queue)
        assign_message.set(message)
        try:
            result = self.assign(*args, **kwargs)
            queue.put(("return", result))
            queue.join()

        except Exception as e:
            logger.exception(e)
            queue.put(("exception", e))
            queue.join()

        janus_queue.set(None)
        assign_message.set(None)


    async def on_assign(self, message: BouncedForwardedAssignMessage):
        queue = janus.Queue()

        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            

            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded,  queue.sync_q, message, args, kwargs)
            queuefut =  self.iterate_queue(queue.async_q, message)

            try:
                await asyncio.gather(threadedfut, queuefut)
                queue.close()
                await queue.wait_closed()

            except asyncio.CancelledError as e:
                await self.log("Received Cancellation for task")

                if not threadedfut.done():
                    await self.log("Sending request to Queue To Cancell")
                    threadedfut.cancel()

                try:
                    await threadedfut
                except asyncio.CancelledError as e:
                    await self.log("Sucessfully Cancelled Thread")
                    raise e

        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            self.console.print_exception()
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))




class FunctionalThreadedGenActor(FunctionalActor):

    def __init__(self, *args, nworkers=4, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.threadpool = ThreadPoolExecutor(nworkers)


    
    async def iterate_queue(self, async_q, message: BouncedForwardedAssignMessage):
        while True:
            val = await async_q.get()
            action = val[0]
            value = val[1]

            if action == "log":
                message = AssignLogMessage(
                    data = {
                        "message": value[0],
                        "level": value[1]
                    },
                    meta = {**message.meta.dict(exclude={"type"})}
                )

                await self.transport.forward(message)
                async_q.task_done()
            if action == "yield":
                await self.transport.forward(AssignYieldsMessage(data={
                    "returns": await shrink_outputs(self.template.node, value) if self.shrink_outputs else value
                }, meta = {**message.meta.dict(exclude={"type"})}
                ))
                async_q.task_done()
            if action == "done":
                await self.transport.forward(AssignDoneMessage(data={
                }, meta = {**message.meta.dict(exclude={"type"})}
                ))
                async_q.task_done()
            if action == "exception":
                await self.transport.forward(AssignCriticalMessage(data={
                        "type": val.__class__.__name__,
                        "message": str(val)
                    }, meta = {**message.meta.dict(exclude={"type"})}
                    ))
                async_q.task_done()
                break  



    def _assign_threaded(self, queue, cancel_event, message, args, kwargs):
        janus_queue.set(queue)
        assign_message.set(message)
        try:
            if cancel_event.is_set(): 
                queue.put(("cancelled", "Cancelled before iteration"))
                queue.join()
                return

            iteration = 0
            for result in self.assign(*args, **kwargs):
                if cancel_event.is_set():
                    queue.put(("cancelled",f"Cancelled after {iteration} Iterations"))
                    queue.join()
                    return

                queue.put(("yield", result))
                queue.join()

            queue.put(("done","Happy doneness"))

        except Exception as e:
            logger.exception(e)
            queue.put(("exception", e))

        janus_queue.set(None)
        assign_message.set(None) 



    async def on_assign(self, message: BouncedForwardedAssignMessage):
        queue = janus.Queue()
        event = threading.Event()
        
        try:
            logger.info("Assigning Number two")
            args, kwargs = await expand_inputs(self.template.node, message.data.args, message.data.kwargs) if self.expand_inputs else (message.data.args, message.data.kwargs)
            

            threadedfut = self.loop.run_in_executor(self.threadpool, self._assign_threaded,  queue.sync_q, event, message, args, kwargs)
            queuefut =  self.iterate_queue(queue.async_q, message)

            try:
                await asyncio.gather(threadedfut, queuefut)
                queue.close()
                await queue.wait_closed()

            except asyncio.CancelledError as e:
                await self.log("Received Cancellation for task")

                if not threadedfut.done():
                    await self.log("Sending request to Queue To Cancell")
                    threadedfut.cancel()

                try:
                    await threadedfut
                except asyncio.CancelledError as e:
                    await self.log("Sucessfully Cancelled Thread")
                    raise e

        except asyncio.CancelledError as e:

            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



        except Exception as e:
            logger.exception(e)
            await self.transport.forward(AssignCriticalMessage(data={
                "type": e.__class__.__name__,
                "message": str(e)
            }, meta = {
                "reference": message.meta.reference,
                "extensions": message.meta.extensions
            }))



from arkitekt.schema import Node
import asyncio

from koil.loop import koil
import re

package_test = re.compile(r"@(?P<package>[^\/]*)\/(?P<interface>[^\/]*)")

async def node_to_action(**kwargs) -> Node:
    node = await Node.asyncs.get(**kwargs)
    return node


def use(**kwargs) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return koil(node_to_action(**kwargs))


async def ause(package=None, interface=None, interactive=False) -> Node:
    """Use a Node on the Platform by Searching for it on its package

    Args:
        package ([type], optional): The package this Node belongs to. Defaults to None.
        interface ([type], optional):  The interface of this Node. Defaults to None.

    Returns:
        Node: The Node
        
    """
    return await node_to_action(package=package, interface=interface)



from abc import abstractmethod
from enum import Enum
from arkitekt.schema.widgets import AllWidgets, QueryWidget, SearchWidget, SliderWidget
from arkitekt.packers.structure import Structure
from typing import Any, Dict, List, Optional, Union
from pydantic import BaseModel, Field, validator
from herre.access.object import GraphQLObject
from arkitekt.packers.registry import get_packer_registry
from typing import ForwardRef
import asyncio




class Port(GraphQLObject):
    key: Optional[str]
    description: Optional[str]
    label: Optional[str]
    transpile: Optional[str]

    @classmethod
    def from_params(cls, widget=None, **kwargs):
        #TODO: This weird widget conversion thing needs to stop, type GraphQLOBject correctly
        port = cls(__typename=cls.__name__, widget= widget.dict() if widget else None, **kwargs) # We ensure creation of a proper object)
        return port


    

class ArgPort(Port):
    widget: Optional[AllWidgets]

    def _repr_html_list(self):
        nana =  f"""
        <div class="container" style="border:1px solid #00000f;padding: 4px;">
            <div class="item item-1 font-xl">{self.key}</div>
            <div class="item item-2">{self.description}</div>
        """
        if self.widget: nana += self.widget._repr_html_list()
        return nana + "</div>"





class IntExpandShrink:

    async def expand(self, value,**kwargs):
        return int(value) if value is not None else None

    async def shrink(self, instance,**kwargs):
       return int(instance) if instance is not None else None


    def to_type(self):
        return int

class BoolExpandShrink:

    async def expand(self, value,**kwargs):
        return bool(value) if value is not None else None

    async def shrink(self, instance,**kwargs):
       return bool(instance) if instance is not None else None

    def to_type(self):
        return bool


class EnumExpandShrink:

    async def expand(self, value,**kwargs):
        return next((key for key, optionvalue in self.options.items() if optionvalue == value), None) if value is not None else None

    async def shrink(self, instance,**kwargs):
       return self.options[instance] if instance else None

    def to_type(self):
        return Enum


class StringExpandShrink:

    async def expand(self, value,**kwargs):
        return str(value) if value is not None else None

    async def shrink(self, instance,**kwargs):
        return str(instance) if instance is not None else None

    def to_type(self):
        return str



class StructureExpandShrink:

    async def expand(self, value, transpile=True):
        if value is None: return None
        structure = get_packer_registry().get_structure(self.identifier)
        return await structure.expand(value)

    async def shrink(self, instance,**kwargs):
        if instance is None: return None
        if isinstance(instance, Structure) or hasattr(instance, "shrink"): return await instance.shrink()
        # Instance we are trying to shrink needs to be transpile to the required model
        from arkitekt.packers.transpilers.registry import get_transpiler_registry
        transpiler = get_transpiler_registry().get_transpiler(instance.__class__.__name__, self.identifier)
        transpiled_instance = await transpiler.transpile(instance)
        return await transpiled_instance.shrink()

    def to_type(self):
        return get_packer_registry().get_structure(self.identifier)
    

class ListExpandShrink:

    async def expand(self, value,**kwargs):
        return await asyncio.gather(*[self.child.expand(item,**kwargs) for item in value]) if value is not None else None

    async def shrink(self, instance,**kwargs):
        assert isinstance(instance, list), f"ListPorts only accept lists! Got {instance}"
        return await asyncio.gather(*[self.child.shrink(item,**kwargs) for item in instance]) if instance is not None else None

    def to_type(self):
        return List[self.child.to_type()]

class DictExpandShrink:

    async def expand(self, value,**kwargs):
        return {key: await self.child.expand(item,**kwargs) for key, item in value.items()} if value is not None else None

    async def shrink(self, instance,**kwargs):
        return {key: await self.child.shrink(item,**kwargs) for key, item in instance.items()} if instance is not None else None

    def to_type(self):
        return Dict[self.child.to_type()]


# Args
class EnumArgPort(ArgPort, EnumExpandShrink):
    options: Optional[dict]
    pass

class IntArgPort(ArgPort, IntExpandShrink):
    pass

class BoolArgPort(ArgPort, BoolExpandShrink):
    pass

class StringArgPort(ArgPort, StringExpandShrink):
    pass


class StructureArgPort(ArgPort, StructureExpandShrink):
    identifier: str


ListArgPort = ForwardRef('ListArgPort')
class ListArgPort(ListExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]


ListArgPort.update_forward_refs()

DictArgPort = ForwardRef('DictArgPort')
class DictArgPort(DictExpandShrink, ArgPort):
    child: Union[IntArgPort, StructureArgPort, StringArgPort, ListArgPort]


   


DictArgPort.update_forward_refs()
#Kwargs

class KwargPort(Port):
    widget: Optional[AllWidgets]

class IntKwargPort(KwargPort, IntExpandShrink):
    default: Optional[int]

class BoolKwargPort(KwargPort, BoolExpandShrink):
    default: Optional[bool]

class EnumKwargPort(KwargPort, EnumExpandShrink):
    default: Any
    options: Optional[dict]
    pass

    @classmethod
    def from_params(cls, widget=None, default=None, **kwargs):
        port = cls(__typename=cls.__name__, widget= widget.dict() if widget else None, default=default._value_, **kwargs) # We ensure creation of a proper object)
        return port

class StringKwargPort(KwargPort, StringExpandShrink):
    default: Optional[str]

class StructureKwargPort(KwargPort, StructureExpandShrink):
    default: Optional[str]
    identifier: str

ListKwargPort = ForwardRef('ListKwargPort')
class ListKwargPort(KwargPort, ListExpandShrink):
    default: Optional[List]
    child: Union[IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort]

ListKwargPort.update_forward_refs()

DictKwargPort = ForwardRef('DictKwargPort')
class DictKwargPort(KwargPort, DictExpandShrink):
    default: Optional[Dict]
    child: Union[IntKwargPort, StructureKwargPort, StringKwargPort, ListKwargPort, DictKwargPort]

DictKwargPort.update_forward_refs()



# Returns

class ReturnPort(Port):
    pass


class StructureReturnPort(ReturnPort, StructureExpandShrink):
    identifier: str

class IntReturnPort(ReturnPort, IntExpandShrink):
    pass

class BoolReturnPort(ReturnPort, BoolExpandShrink):
    pass

class StringReturnPort(ReturnPort, StringExpandShrink):
    pass

ListReturnPort = ForwardRef('ListReturnPort')
class ListReturnPort(ReturnPort, ListExpandShrink):
    child: Union[IntReturnPort, StructureReturnPort, StringReturnPort, ListReturnPort]

ListReturnPort.update_forward_refs()


DictReturnPort = ForwardRef('DictReturnPort')
class DictReturnPort(ReturnPort, DictExpandShrink):
    child: Union[IntReturnPort, StructureReturnPort, StringReturnPort, ListReturnPort, DictReturnPort]

DictReturnPort.update_forward_refs()


AllArgPort = Union[StructureArgPort, StringArgPort, IntArgPort, ListArgPort, DictArgPort, EnumArgPort, BoolArgPort]
AllKwargPort = Union[StructureKwargPort, StringKwargPort, IntKwargPort, ListKwargPort, DictKwargPort, EnumKwargPort, BoolKwargPort]
AllReturnPort = Union[StructureReturnPort, StringReturnPort, IntReturnPort, ListReturnPort, DictReturnPort, BoolReturnPort]






from abc import abstractmethod
from arkitekt.agents.base import Agent
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage




import logging

logger = logging.getLogger(__name__)


class StandardAgent(Agent):

    @abstractmethod
    async def on_bounced_provide(message: BouncedProvideMessage):
        raise NotImplementedError()

    @abstractmethod
    async def on_bounced_unprovide(message: BouncedUnprovideMessage):
        raise NotImplementedError()


    @abstractmethod
    async def on_bounced_assign(message: BouncedForwardedAssignMessage):
        raise NotImplementedError()


    @abstractmethod
    async def on_bounced_unassign(message: BouncedForwardedUnassignMessage):
        raise NotImplementedError()


    
    async def handle_bounced_provide(self, message: BouncedProvideMessage):
        try:
            await self.on_bounced_provide(message)
            await self.transport.forward(ProvideLogMessage.from_log(message.meta.reference, "Forwarded to Agent", extensions=message.meta.extensions))
        except Exception as e:
            logger.error(e)
            await self.transport.forward(ProvideTransitionMessage.from_critical(e, message.meta.reference, extensions=message.meta.extensions))
            


    async def handle_bounced_unprovide(self, message: BouncedUnprovideMessage):
        try:
            await self.on_bounced_unprovide(message)
            await self.transport.forward(ProvideLogMessage.from_log(message.data.provision, "Actor Delation Happening", extensions=message.meta.extensions))
        except Exception as e:
            logger.error(e)
            await self.transport.forward(ProvideTransitionMessage.from_critical(e, message.data.provision, extensions=message.meta.extensions))
     

    async def handle_bounced_assign(self, message: BouncedForwardedAssignMessage):
        try:
            await self.on_bounced_assign(message)
            await self.transport.forward(AssignLogMessage.from_log(message.meta.reference, "Forwared Assignment", extensions=message.meta.extensions))

        except Exception as e:
            await self.transport.forward(AssignCriticalMessage.from_critical(message.meta.reference, e, extensions=message.meta.extensions))
            raise e

    async def handle_bounced_unassign(self, message: BouncedForwardedUnassignMessage):
        try:
            await self.on_bounced_unassign(message)
            await self.transport.forward(AssignLogMessage.from_log(message.data.assignation, "Received Cancellation", extensions=message.meta.extensions))


        except Exception as e:
            logger.error(e)
            await self.transport.forward(AssignCriticalMessage.from_critical(message.data.assignation, e, extensions=message.meta.extensions))
            raise e




from qtpy import QtCore, QtGui, QtWidgets
from arkitekt.agents.base import Agent
from arkitekt.agents.qt import QtAgent
from arkitekt.config import ArkitektConfig
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template
import webbrowser

def open_in_arkitekt(path):
    config = ArkitektConfig.from_konfik()
    webbrowser.open(f"http://{config.host}:3000{path}")



class PortsWrapped(QtWidgets.QWidget):

    def __init__(self, ports, title, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QVBoxLayout()
        self.formGroupBox = QtWidgets.QGroupBox(title)
        qlayout = QtWidgets.QFormLayout()

        for port in ports:
            argLabel = QtWidgets.QLabel(f"{port.typename} | {port.key}")
            argDescription = QtWidgets.QLabel(port.description)
            argDescription.setFont(QtGui.QFont("Arial", 8))
            qlayout.addWidget(argLabel)
            qlayout.addWidget(argDescription)

        
        self.formGroupBox.setLayout(qlayout)
        self.layout.addWidget(self.formGroupBox)
        self.setLayout(self.layout)







class TemplateDetailWidget(QtWidgets.QWidget):

    def __init__(self, template: Template, agent: Agent, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.template = template
        self.layout = QtWidgets.QVBoxLayout()

        name = QtWidgets.QLabel(template.node.name)
        name.setFont(QtGui.QFont("Arial", 16))

        identifier = QtWidgets.QLabel(f"@{template.node.package}/{template.node.interface}")
        identifier.setFont(QtGui.QFont("Arial", 10))

        description = QtWidgets.QLabel(template.node.description)
        description.setWordWrap(True)

        open_button = QtWidgets.QPushButton("Open in Arkitekt")
        open_button.clicked.connect(self.open_in_arkitekt)



        self.layout.addWidget(name)
        self.layout.addWidget(identifier)
        self.layout.addWidget(description)


        self.layout.addWidget(PortsWrapped(template.node.args, "Arguments"))
        self.layout.addWidget(PortsWrapped(template.node.kwargs, "Constants"))
        self.layout.addWidget(PortsWrapped(template.node.returns, "Returns"))

        for kwarg in template.node.kwargs:
            kwargLabel = QtWidgets.QLabel(kwarg.typename + " | " + kwarg.key + " | " + str(kwarg.default))
            self.layout.addWidget(kwargLabel)



        self.layout.addWidget(open_button)
        self.setLayout(self.layout)

    def open_in_arkitekt(self):
        open_in_arkitekt(f"/node/{self.template.node.id}")






class TemplatesListItemWidget(QtWidgets.QWidget):

    def __init__(self, template: Template, agent: Agent, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.template = template
        self.agent = agent

        self.row = QtWidgets.QHBoxLayout()

        left = QtWidgets.QWidget()
        leftlayout = QtWidgets.QVBoxLayout()

        self.dialog = TemplateDetailWidget(self.template, self.agent)


        node_label = QtWidgets.QLabel(template.node.name)
        node_label.setFont(QtGui.QFont("Arial", 10))

        context_label = QtWidgets.QLabel(f"@{template.node.package}/{template.node.interface}")
        context_label.setFont(QtGui.QFont("Arial", 8))

        leftlayout.addWidget(node_label)
        leftlayout.addWidget(context_label)

        left.setLayout(leftlayout)

        open_button = QtWidgets.QToolButton()
        open_button.clicked.connect(self.open_provision)

        self.row.addWidget(left)
        self.row.addWidget(open_button,  alignment=QtCore.Qt.AlignRight)

        self.setLayout(self.row)

    def open_provision(self):
        self.dialog.show()




class TemplatesWidget(QtWidgets.QWidget):

    def __init__(self, qt_agent: QtAgent = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.agent = qt_agent
        self.layout = QtWidgets.QVBoxLayout()
        self.listWidget = QtWidgets.QListWidget()
        self.layout.addWidget(self.listWidget)


        self.agent.provide_signal.connect(self.on_provide)

        self.setLayout(self.layout)
        self.provisions = {}


    def on_provide(self, provide_state):
        print("Hidden Here")
        if provide_state:
            self.listWidget.clear()
            for key, template in self.agent.templateTemplatesMap.items():
                item = QtWidgets.QListWidgetItem()
                w = TemplatesListItemWidget(template, self.agent)
                item.setSizeHint(w.minimumSizeHint())
                self.listWidget.addItem(item)
                self.listWidget.setItemWidget(item, w)

        else:
            self.listWidget.clear()





from enum import Enum
from ..base import MessageDataModel


class LogLevel(str, Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"
    WARNING = "WARNING"
    WARN = "WARN"


class LogDataModel(MessageDataModel):
    level: LogLevel
    message: str

    @classmethod
    def from_log(cls, message, level=LogLevel.INFO):
        return cls(level=level, message=message)




from typing import Dict
from arkitekt.actors.actify import define
from arkitekt.agents.app import AppAgent
from arkitekt.agents.qt.actor import QtActor
from qtpy.QtCore import QObject, Signal
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from arkitekt.agents.base import Agent, AgentException
import logging
import uuid

logger = logging.getLogger(__name__)


class AgentSignals(QObject):
    provide = Signal(BouncedProvideMessage)
    unprovide = Signal(BouncedUnprovideMessage)
    provide_transition = Signal(ProvideTransitionMessage)


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


class QtAgent(AppAgent, QObject):
    provide_signal = Signal(bool)
    provision_signal = Signal(BouncedProvideMessage)
    unprovision_signal = Signal(BouncedUnprovideMessage)

    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self,*args, strict=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.strict = strict
        self.assignFutures = {}
        self.provideFutures = {}
        self.unprovideFutures = {}
        self.appWorkers = {}


    def on_task_done(self, future):
        print(future)

    async def on_bounced_provide(self, message: BouncedProvideMessage):
        if message.data.template in self.templateActorsMap:
            if message.meta.reference in self.templateActorsMap:
                if self.strict: raise AgentException("Already Running Provision Received Again. Right now causing Error. Might be omitted")
                again_provided = ProvideTransitionMessage(data={
                    "message": "Provision was running on this Instance. Probably a freaking race condition",
                    "state": ProvideState.ACTIVE
                    }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
                await self.transport.forward(again_provided)
            else:
                actor = self.templateActorsMap[message.data.template]
                self.runningActors[message.meta.reference] = actor
                print("RUnning Actor Here")
                task = self.loop.create_task(actor.arun(message, self))
                task.add_done_callback(self.on_task_done)
                self.runningTasks[message.meta.reference] = task

        else:
            raise AgentException("No approved actors for this template")


    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        if message.data.provision not in self.runningActors: raise AgentException("Already Running Provision Received Again. Right now causing Error. Might be omitted")
        actor = self.runningActors[message.data.provision]
        
        logger.info(f"Cancelling {actor}")
        self.runningTasks[message.data.provision].cancel()

    async def on_bounced_assign(self, message: BouncedForwardedAssignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)    
        else:
            if self.strict: raise AgentException("Received Assignment for not running Provision")

    async def on_bounced_unassign(self, message: BouncedForwardedUnassignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)
            
        else:
            if self.strict: raise AgentException("Received Assignment for not running Provision")
            logger.info("We didnt have this assignment, setting Cancellation anyways")
            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": "Fake Cancellation trough Provider"
            }, meta = {
                "reference": message.data.assignation
            }))
                


    def register(self, function_or_node, widgets={}, transpilers: Dict[str, Transpiler] = None, on_provide = None, on_unprovide = None, on_assign = None, timeout=500, **params) -> QtActor:

        # Simple bypass for now
        defined_actor = QtActor(qt_assign=on_assign, qt_on_provide=on_provide, qt_on_unprovide=on_unprovide, loop=self.loop, timeout=timeout, **params)
        
        if isinstance(function_or_node, str):
            self.templatedUnqueriedNodes.append(({"q": function_or_node}, defined_actor, params))
           
        if isinstance(function_or_node, Node):
            self.templatedNodes.append((function_or_node, defined_actor, params))

        else:
            defined_node = define(function=function_or_node, widgets=widgets)
            self.templatedNewNodes.append((defined_node, defined_actor, params))

        return defined_actor

    async def approve_nodes_and_templates(self):
        await super().approve_nodes_and_templates()
        self.provide_signal.emit(True)
        return 

    async def aprovide(self):
        try:
            await super().aprovide()
        except asyncio.CancelledError as e:
            self.provide_signal.emit(False)
            raise e


    
   

    


from qtpy import QtCore, QtGui
from qtpy import QtWidgets
from arkitekt.agents.base import Agent
from arkitekt.agents.qt import QtAgent
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template



class ProvisionDetailWidget(QtWidgets.QWidget):

    def __init__(self, provide: BouncedProvideMessage, agent: Agent, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        template: Template = agent.templateTemplatesMap[provide.data.template]
        self.layout = QtWidgets.QVBoxLayout()

        name = QtWidgets.QLabel(template.node.name)
        name.setFont(QtGui.QFont("Arial", 16))

        identifier = QtWidgets.QLabel(f"@{template.node.package}/{template.node.interface}")
        identifier.setFont(QtGui.QFont("Arial", 10))

        description = QtWidgets.QLabel(template.node.description)
        description.setWordWrap(True)

        self.layout.addWidget(name)
        self.layout.addWidget(identifier)
        self.layout.addWidget(description)
        self.setLayout(self.layout)




class ProvisionListWidget(QtWidgets.QWidget):

    def __init__(self, provide: BouncedProvideMessage, agent: Agent, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        template: Template = agent.templateTemplatesMap[provide.data.template]
        self.provide = provide
        self.agent = agent

        self.row = QtWidgets.QHBoxLayout()

        left = QtWidgets.QWidget()
        leftlayout = QtWidgets.QVBoxLayout()

        self.dialog = ProvisionDetailWidget(self.provide, self.agent)


        node_label = QtWidgets.QLabel(template.node.name)
        node_label.setFont(QtGui.QFont("Arial", 10))

        context_label = QtWidgets.QLabel(provide.meta.context.user)
        context_label.setFont(QtGui.QFont("Arial", 8))

        leftlayout.addWidget(node_label)
        leftlayout.addWidget(context_label)

        left.setLayout(leftlayout)

        open_button = QtWidgets.QToolButton()
        open_button.clicked.connect(self.open_provision)

        self.row.addWidget(left)
        self.row.addWidget(open_button,  alignment=QtCore.Qt.AlignRight)

        self.setLayout(self.row)

    def open_provision(self):
        self.dialog.show()




class ProvisionsWidget(QtWidgets.QWidget):

    def __init__(self, qt_agent: QtAgent = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.agent = qt_agent
        self.layout = QtWidgets.QVBoxLayout()
        self.listWidget = QtWidgets.QListWidget()
        self.layout.addWidget(self.listWidget)

        self.setLayout(self.layout)
        self.provisions = {}

        self.agent.provision_signal.connect(self.provision_in)

    def provision_in(self, provide: BouncedProvideMessage):
        self.provisions[provide.meta.reference] = provide

        self.listWidget.clear()
        print(self.provisions)
        for key, provide in self.provisions.items():
            item = QtWidgets.QListWidgetItem()
            w = ProvisionListWidget(provide, self.agent)
            item.setSizeHint(w.minimumSizeHint())
            self.listWidget.addItem(item)
            self.listWidget.setItemWidget(item, w)






from arkitekt.agents.standard import StandardAgent
from arkitekt.messages.postman.assign.assign_cancelled import AssignCancelledMessage
from arkitekt.messages.postman.unassign.bounced_forwarded_unassign import BouncedForwardedUnassignMessage
from arkitekt.messages.postman.provide.provide_transition import ProvideState, ProvideTransitionMessage
from arkitekt.messages.postman.assign.assign_log import AssignLogMessage
from arkitekt.messages.postman.assign.assign_critical import AssignCriticalMessage
from arkitekt.messages.postman.assign.bounced_forwarded_assign import BouncedForwardedAssignMessage
from arkitekt.messages.postman.assign.bounced_assign import BouncedAssignMessage
from arkitekt.messages.postman.provide.provide_critical import ProvideCriticalMessage
from arkitekt.messages.postman.log import LogLevel
from arkitekt.messages.postman.provide.provide_log import ProvideLogMessage
import asyncio
from herre.wards.base import WardException
from arkitekt.actors.actify import actify, define
from arkitekt.actors.base import Actor
from arkitekt.messages.postman.unprovide.bounced_unprovide import BouncedUnprovideMessage
from arkitekt.messages.base import MessageDataModel, MessageModel
from arkitekt.messages.postman.provide.bounced_provide import BouncedProvideMessage
from arkitekt.schema.template import Template
from arkitekt.schema.node import Node
from arkitekt.packers.transpilers import Transpiler
from typing import Callable, Dict, List, Tuple, Type
from arkitekt.agents.base import Agent, AgentException, parse_params
import logging
from herre.console import get_current_console

logger = logging.getLogger(__name__)


class AppAgent(StandardAgent):
    ACTOR_PENDING_MESSAGE = "Actor is Pending"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        # Running Actors indexed by their ID
        self.runningActors: Dict[str, Actor] = {}
        self.runningTasks: Dict[str, asyncio.Task] = {}

        self.templatedUnqueriedNodes:  List[Tuple[dict, Callable]] = [] # dict are queryparams for the node
        self.templatedNodes: List[Tuple[Node, Callable]] = [] # node is already saved and has id
        self.templatedNewNodes: List[Tuple[Node, Callable]] = [] # Node is not saved and has undefined id


        self.approvedTemplates: List[Tuple[Template, Callable]] = [] # Template is approved 


        # IMportant Maps
        self.templateActorsMap = {}
        self.templateTemplatesMap = {}



    async def on_transport_about_to_connect(self):
        await self.approve_nodes_and_templates()
        return 

    async def on_transport_connected(self):
        print(f"Hosting {self.templateActorsMap.keys()}")


    def on_task_done(self, future):
        print(future)


    async def on_bounced_provide(self, message: BouncedProvideMessage):
        
        if message.data.template in self.templateActorsMap:
            if message.meta.reference not in self.runningActors:
                # Didn not exist before
                actor = self.templateActorsMap[message.data.template] # creating out little Actor
                self.runningActors[message.meta.reference] = actor
                task = self.loop.create_task(actor.arun(message, self))
                task.add_done_callback(self.on_task_done)
                self.runningTasks[message.meta.reference] = task
                
            else:
                if self.strict: raise AgentException("Already Running Provision Received Again. Right now causing Error. Might be omitted")
                again_provided = ProvideTransitionMessage(data={
                    "message": "Provision was running on this Instance. Probably a freaking race condition",
                    "state": ProvideState.ACTIVE
                    }, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
                await self.transport.forward(again_provided)

        else:
            raise AgentException("No approved actors for this template")


    async def on_bounced_unprovide(self, message: BouncedUnprovideMessage):
        if message.data.provision not in self.runningActors: raise AgentException("Already Running Provision Received Again. Right now causing Error. Might be omitted")
        actor = self.runningActors[message.data.provision]
        
        logger.info(f"Cancelling {actor}")
        self.runningTasks[message.data.provision].cancel()


    async def on_bounced_assign(self, message: BouncedForwardedAssignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)    
        else:
            if self.strict: raise AgentException("Received Assignment for not running Provision")

    async def on_bounced_unassign(self, message: BouncedForwardedUnassignMessage):

        if message.data.provision in self.runningActors:
            actor = self.runningActors[message.data.provision]
            await actor.acall(message=message)
            
        else:
            if self.strict: raise AgentException("Received Assignment for not running Provision")
            logger.info("We didnt have this assignment, setting Cancellation anyways")
            await self.transport.forward(AssignCancelledMessage(data={
                "canceller": "Fake Cancellation trough Provider"
            }, meta = {
                "reference": message.data.assignation
            }))
                




    async def approve_nodes_and_templates(self):

        if self.templatedUnqueriedNodes:
            for query_params, defined_actor, params in self.templatedUnqueriedNodes:
                try:
                    arkitekt_node = await Node.asyncs.get(**query_params)
                    self.templatedNodes.append((arkitekt_node, defined_actor, params))
                except WardException as e:
                    logger.exception(e)
                    if self.strict: raise AgentException(f"Couldn't find Node for query {query_params}") from e   

        if self.templatedNewNodes:
            for defined_node, defined_actor, params in self.templatedNewNodes:
                # Defined Node are nodes that are not yet reflected on arkitekt (i.e they dont have an instance
                # id so we are trying to send them to arkitekt)
                try:
                    arkitekt_node = await Node.asyncs.create(**defined_node.dict(as_input=True))
                    self.templatedNodes.append((arkitekt_node, defined_actor, params))   
                except WardException as e:
                    logger.exception(e)
                    if self.strict: raise AgentException(f"Couldn't create Node for defintion {defined_node}") from e

        if self.templatedNodes:
            # This is an arkitekt Node and we can generate potential Templates
            for arkitekt_node, defined_actor, params in self.templatedNodes:
                try:
                    params = await parse_params(params) # Parse the parameters for template creation
                    arkitekt_template = await Template.asyncs.create(node=arkitekt_node, params=params)
                    self.approvedTemplates.append((arkitekt_template, defined_actor, params))  
                except WardException as e:
                    logger.exception(e)
                    if self.strict: raise AgentException(f"Couldn't approve template for node {arkitekt_node}") from e

        if self.approvedTemplates:

            for arkitekt_template, defined_actor, params in self.approvedTemplates:

                # Generating Maps for Easy access
                self.templateActorsMap[arkitekt_template.id] = defined_actor
                self.templateTemplatesMap[arkitekt_template.id] = arkitekt_template

                if self.panel: self.panel.add_to_actor_map(arkitekt_template, defined_actor) 



    def register(self, *args, widgets={}, transpilers: Dict[str, Transpiler] = None, on_provide = None, on_unprovide = None, **params):

        def real_decorator(function):
            # Simple bypass for now
            def wrapped_function(*args, **kwargs):
                return function(*args, **kwargs)

            
            defined_actor = actify(function, on_provide=on_provide, on_unprovide=on_unprovide, **params)

            if len(args) == 0:
                defined_node = define(function=function, widgets=widgets)
                self.templatedNewNodes.append((defined_node, defined_actor, params))

            if len(args) == 1:
                if isinstance(args[0], str):
                    self.templatedUnqueriedNodes.append(({"q": args[0]}, defined_actor, params))

                if isinstance(args[0], Node):
                    self.templatedNodes.append((args[0], defined_actor, params))

                # We are registering this as a template

            return wrapped_function

    
        return real_decorator
